"use strict";

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

Object.defineProperty(exports, "__esModule", { value: true });
var cloneDeep = require("lodash.clonedeep");
var global_state_1 = require("./global-state");
/**
 * 是否是基本类型
 */
function isPrimitive(value) {
    if (value === null || value === undefined) {
        return true;
    }
    if (typeof value === "string" || typeof value === "number" || typeof value === "boolean" || value instanceof Date) {
        return true;
    }
    return false;
}
exports.isPrimitive = isPrimitive;
/**
 * redux thunk
 */
exports.createThunkMiddleware = function (_ref) {
    var dispatch = _ref.dispatch,
        getState = _ref.getState;
    return function (next) {
        return function (action) {
            if (typeof action === "function") {
                return action(dispatch, getState);
            }
            return next(action);
        };
    };
};
/**
 * 空函数
 */
// tslint:disable-next-line:no-empty
exports.noop = function () {};
/**
 * 是否在 action 中
 */
function inAction() {
    return global_state_1.globalState.batchDeep !== 0;
}
exports.inAction = inAction;
/**
 * 是否在 track 中
 */
function inTrack() {
    return !!global_state_1.globalState.currentReaction;
}
exports.inTrack = inTrack;
/**
 * 找到某个对象、键值的 binder
 */
function getBinder(object, key) {
    var keysForObject = global_state_1.globalState.objectReactionBindings.get(object);
    if (!keysForObject) {
        keysForObject = new Map();
        global_state_1.globalState.objectReactionBindings.set(object, keysForObject);
    }
    var reactionsForKey = keysForObject.get(key);
    if (!reactionsForKey) {
        reactionsForKey = new Set();
        keysForObject.set(key, reactionsForKey);
    }
    return {
        binder: keysForObject,
        keyBinder: reactionsForKey
    };
}
exports.getBinder = getBinder;
/**
 * 开启调试模式
 */
function startDebug() {
    global_state_1.globalState.useDebug = true;
}
exports.startDebug = startDebug;
/**
 * 关闭调试模式
 */
function stopDebug() {
    global_state_1.globalState.useDebug = false;
}
exports.stopDebug = stopDebug;
/**
 * 注册 parentInfo
 */
function registerParentInfo(target, key, value) {
    if (value !== null && (typeof value === "undefined" ? "undefined" : _typeof(value)) === "object") {
        global_state_1.globalState.parentInfo.set(value, {
            parent: target,
            key: key
        });
    }
}
exports.registerParentInfo = registerParentInfo;
/**
 * debug 入栈 action
 */
function debugInAction(actionName) {
    var debugOutputBundleAction = {
        name: actionName,
        changeList: [],
        type: "action"
    };
    global_state_1.globalState.debugOutputActionMapBatchDeep.set(global_state_1.globalState.batchDeep, debugOutputBundleAction);
    global_state_1.globalState.currentDebugOutputAction = debugOutputBundleAction;
    // 如果深度等于 1，生成唯一 id 给这个 action
    if (global_state_1.globalState.batchDeep === 1) {
        debugOutputBundleAction.id = getUniqueId();
        global_state_1.globalState.currentDebugId = debugOutputBundleAction.id;
    }
    // 如果当前深度大于 1，就作为加到父级的 changeList
    if (global_state_1.globalState.batchDeep > 1) {
        global_state_1.globalState.debugOutputActionMapBatchDeep.get(global_state_1.globalState.batchDeep - 1).changeList.push({
            type: "action",
            action: debugOutputBundleAction
        });
    }
}
exports.debugInAction = debugInAction;
/**
 * debug 出栈 action
 */
function debugOutAction() {
    // 每次出栈，都要更新 currentDebugOutputAction
    global_state_1.globalState.currentDebugOutputAction = global_state_1.globalState.debugOutputActionMapBatchDeep.get(global_state_1.globalState.batchDeep);
    if (!inAction()) {
        var cloneDebugInfo = null;
        try {
            cloneDebugInfo = JSON.parse(JSON.stringify(global_state_1.globalState.debugOutputActionMapBatchDeep.get(1)));
        } catch (error) {
            return;
        }
        // 如果完全出队列了，把存储的 debug 信息输出给 debug 事件，并清空 debug 信息
        global_state_1.globalState.event.emit("debug", cloneDebugInfo);
        global_state_1.globalState.currentDebugOutputAction = null;
        global_state_1.globalState.debugOutputActionMapBatchDeep.clear();
        // 此时不能清空 globalState.currentDebugId = null，因为后续要传给 callback
        // action 与 debug 调用顺序：startBatch -> debugInAction -> ...multiple nested startBatch and endBatch -> debugOutAction -> reaction -> observe
    }
}
exports.debugOutAction = debugOutAction;
/**
 * 获取对象路径
 */
function getCallStack(target) {
    var callStack = [];
    if (!global_state_1.globalState.parentInfo.has(target)) {
        callStack.unshift(target.constructor.name);
    } else {
        var currentTarget = target;
        var runCount = 0;
        while (global_state_1.globalState.parentInfo.has(currentTarget)) {
            var parentInfo = global_state_1.globalState.parentInfo.get(currentTarget);
            // 添加调用队列
            callStack.unshift(parentInfo.key);
            // 如果父级没有父级了，给调用队列添加父级名称
            if (!global_state_1.globalState.parentInfo.has(parentInfo.parent)) {
                callStack.unshift(parentInfo.parent.constructor.name);
            }
            currentTarget = parentInfo.parent;
            runCount++;
            if (runCount >= 50) {
                break;
            }
        }
    }
    return callStack;
}
/**
 * 打印 diff 路径
 */
function printDiff(target, key, oldValue, value) {
    var callStack = getCallStack(target);
    reportChange({
        type: "change",
        callStack: callStack,
        oldValue: cloneDeep(oldValue),
        key: key,
        value: cloneDeep(value)
    });
}
exports.printDiff = printDiff;
/**
 * 打印删除路径
 */
function printDelete(target, key) {
    var callStack = getCallStack(target);
    reportChange({
        type: "delete",
        callStack: callStack,
        key: key
    });
}
exports.printDelete = printDelete;
/**
 * 自定义打印
 */
function printCustom(target) {
    for (var _len = arguments.length, customMessage = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        customMessage[_key - 1] = arguments[_key];
    }

    var callStack = getCallStack(target);
    reportChange({
        type: "custom",
        callStack: callStack,
        customMessage: customMessage
    });
}
exports.printCustom = printCustom;
/**
 * 通知修改
 */
function reportChange(change) {
    if (global_state_1.globalState.currentDebugOutputAction) {
        global_state_1.globalState.currentDebugOutputAction.changeList.push(change);
    } else {
        global_state_1.globalState.event.emit("debug", {
            id: getUniqueId(),
            name: null,
            changeList: [change],
            type: "isolated"
        });
    }
}
/**
 * 设置处于严格模式
 */
function useStrict() {
    global_state_1.globalState.strictMode = true;
}
exports.useStrict = useStrict;
/**
 * 取消严格模式
 */
function cancelStrict() {
    global_state_1.globalState.strictMode = false;
}
exports.cancelStrict = cancelStrict;
/**
 * 生成随机唯一 id
 */
function getUniqueId() {
    return global_state_1.globalState.uniqueIdCounter++;
}
exports.getUniqueId = getUniqueId;
//# sourceMappingURL=utils.js.map